#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

import warnings

from dataclasses import dataclass
from io import BufferedWriter
from pathlib import Path
from typing import Dict, Optional
import os
from requests import Response
from requests.structures import CaseInsensitiveDict

from aistore.sdk.archive_config import ArchiveConfig
from aistore.sdk.blob_download_config import BlobDownloadConfig
from aistore.sdk.etl import ETLConfig
from aistore.sdk.const import (
    BYTE_RANGE_PREFIX_LENGTH,
    DEFAULT_CHUNK_SIZE,
    HTTP_METHOD_DELETE,
    HTTP_METHOD_HEAD,
    QPARAM_ARCHPATH,
    QPARAM_ARCHREGX,
    QPARAM_ARCHMODE,
    QPARAM_ETL_NAME,
    QPARAM_ETL_ARGS,
    QPARAM_LATEST,
    ACT_PROMOTE,
    HTTP_METHOD_POST,
    URL_PATH_OBJECTS,
    HEADER_RANGE,
    ACT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_DOWNLOAD,
    HEADER_OBJECT_BLOB_WORKERS,
    HEADER_OBJECT_BLOB_CHUNK_SIZE,
)
from aistore.sdk.provider import Provider
from aistore.sdk.obj.object_client import ObjectClient
from aistore.sdk.obj.object_reader import ObjectReader
from aistore.sdk.obj.object_writer import ObjectWriter
from aistore.sdk.request_client import RequestClient
from aistore.sdk.types import (
    ActionMsg,
    PromoteAPIArgs,
    BlobMsg,
)
from aistore.sdk.obj.object_props import ObjectProps


@dataclass
class BucketDetails:
    """
    Metadata about a bucket, used by objects within that bucket.
    """

    name: str
    provider: Provider
    qparams: Dict[str, str]
    path: str


class Object:
    """
    Provides methods for interacting with an object in AIS.

    Args:
        client (RequestClient): Client used for all http requests.
        bck_details (BucketDetails): Metadata about the bucket to which this object belongs.
        name (str): Name of the object.
        props (ObjectProps, optional): Properties of the object, as updated by head(), optionally pre-initialized.
    """

    def __init__(
        self,
        client: RequestClient,
        bck_details: BucketDetails,
        name: str,
        props: ObjectProps = None,
    ):
        self._client = client
        self._bck_details = bck_details
        self._bck_path = f"{URL_PATH_OBJECTS}/{ bck_details.name}"
        self._object_path = f"{self._bck_path}/{name}"
        self._name = name
        self._props = props

    @property
    def bucket_name(self) -> str:
        """Name of the bucket where this object resides."""
        return self._bck_details.name

    @property
    def bucket_provider(self) -> Provider:
        """Provider of the bucket where this object resides (e.g. ais, s3, gcp)."""
        return self._bck_details.provider

    @property
    def query_params(self) -> Dict[str, str]:
        """Query params used as a base for constructing all requests for this object."""
        return self._bck_details.qparams

    @property
    def name(self) -> str:
        """Name of this object."""
        return self._name

    @property
    def props(self) -> ObjectProps:
        """Properties of this object."""
        return self._props

    def head(self) -> CaseInsensitiveDict:
        """
        Requests object properties and returns headers. Updates props.

        Returns:
            Response header with the object properties.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        headers = self._client.request(
            HTTP_METHOD_HEAD,
            path=self._object_path,
            params=self.query_params,
        ).headers
        self._props = ObjectProps(headers)
        return headers

    # pylint: disable=too-many-arguments,too-many-locals
    def get_reader(
        self,
        archive_config: Optional[ArchiveConfig] = None,
        blob_download_config: Optional[BlobDownloadConfig] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: Optional[ETLConfig] = None,
        writer: Optional[BufferedWriter] = None,
        latest: bool = False,
        byte_range: Optional[str] = None,
        direct: bool = False,
    ) -> ObjectReader:
        """
        Creates and returns an ObjectReader with access to object contents
        and optionally writes to a provided writer.

        Args:
            archive_config (Optional[ArchiveConfig]): Settings for archive extraction.
            blob_download_config (Optional[BlobDownloadConfig]): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (Optional[ETLConfig]): Settings for ETL-specific operations (name, args).
            writer (Optional[BufferedWriter]): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (Optional[str]): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.
            direct (bool, optional): If True, the object content is read directly from the target node,
                bypassing the proxy.

        Returns:
            ObjectReader: An iterator for streaming object content.

        Raises:
            ValueError: If Byte Range is used with Blob Download.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """

        params = self.query_params.copy()
        headers = {}
        byte_range_tuple = None

        # Archive Configuration
        if archive_config:
            if archive_config.mode:
                params[QPARAM_ARCHMODE] = archive_config.mode.value
            params.update(
                {
                    QPARAM_ARCHPATH: archive_config.archpath,
                    QPARAM_ARCHREGX: archive_config.regex,
                }
            )

        # Blob Download Configuration
        if blob_download_config:
            headers.update(
                {
                    HEADER_OBJECT_BLOB_DOWNLOAD: "true",
                    HEADER_OBJECT_BLOB_CHUNK_SIZE: blob_download_config.chunk_size,
                    HEADER_OBJECT_BLOB_WORKERS: blob_download_config.num_workers,
                }
            )

        # ETL Configuration
        if etl:
            params[QPARAM_ETL_NAME] = etl.name
            if etl.args:
                params[QPARAM_ETL_ARGS] = etl.args

        # Latest Object Version
        if latest:
            params[QPARAM_LATEST] = "true"

        # Byte Range Validation
        if byte_range and blob_download_config:
            raise ValueError("Cannot use Byte Range with Blob Download.")

        if byte_range:
            # For range formatting, see the spec:
            # https://www.rfc-editor.org/rfc/rfc7233#section-2.1
            headers = {HEADER_RANGE: byte_range}
            # Extract left (range_l) and right (range_r) bounds from the byte range string
            headers[HEADER_RANGE] = byte_range
            byte_range_l, _, byte_range_r = byte_range[
                BYTE_RANGE_PREFIX_LENGTH:
            ].partition("-")
            byte_range_tuple = (
                int(byte_range_l) if byte_range_l else None,
                int(byte_range_r) if byte_range_r else None,
            )

        # Object Client
        obj_client = ObjectClient(
            request_client=self._client,
            path=self._object_path,
            params=params,
            headers=headers,
            byte_range=byte_range_tuple,
            uname=os.path.join(self._bck_details.path, self.name) if direct else None,
        )

        obj_reader = ObjectReader(object_client=obj_client, chunk_size=chunk_size)

        if writer:
            writer.writelines(obj_reader)

        return obj_reader

    # pylint: disable=too-many-arguments
    def get(
        self,
        archive_config: ArchiveConfig = None,
        blob_download_config: BlobDownloadConfig = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        etl: ETLConfig = None,
        writer: BufferedWriter = None,
        latest: bool = False,
        byte_range: str = None,
    ) -> ObjectReader:
        """
        Deprecated: Use 'get_reader' instead.

        Creates and returns an ObjectReader with access to object contents and optionally writes to a provided writer.

        Args:
            archive_config (ArchiveConfig, optional): Settings for archive extraction.
            blob_download_config (BlobDownloadConfig, optional): Settings for using blob download.
            chunk_size (int, optional): Chunk size to use while reading from stream.
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).
            writer (BufferedWriter, optional): User-provided writer for writing content output.
                The user is responsible for closing the writer.
            latest (bool, optional): GET the latest object version from the associated remote bucket.
            byte_range (str, optional): Byte range in RFC 7233 format for single-range requests
                (e.g., "bytes=0-499", "bytes=500-", "bytes=-500").
                See: https://www.rfc-editor.org/rfc/rfc7233#section-2.1.

        Returns:
            ObjectReader: An ObjectReader that can be iterated over to stream chunks of object content
            or used to read all content directly.

        Raises:
            ValueError: If Byte Range is used with Blob Download.
            requests.RequestException: If an error occurs during the request.
            requests.ConnectionError: If there is a connection error.
            requests.ConnectionTimeout: If the connection times out.
            requests.ReadTimeout: If the read operation times out.
        """
        warnings.warn(
            "The 'get' method is deprecated and will be removed in a future release. "
            "Please use 'get_reader' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_reader(
            archive_config=archive_config,
            blob_download_config=blob_download_config,
            chunk_size=chunk_size,
            etl=etl,
            writer=writer,
            latest=latest,
            byte_range=byte_range,
        )

    def get_semantic_url(self) -> str:
        """
        Get the semantic URL to the object

        Returns:
            Semantic URL to get object
        """

        return f"{self.bucket_provider.value}://{self.bucket_name}/{self._name}"

    def get_url(self, archpath: str = "", etl: ETLConfig = None) -> str:
        """
        Get the full url to the object including base url and any query parameters

        Args:
            archpath (str, optional): If the object is an archive, use `archpath` to extract a single file
                from the archive
            etl (ETLConfig, optional): Settings for ETL-specific operations (name, meta).

        Returns:
            Full URL to get object

        """
        params = self.query_params.copy()
        if archpath:
            params[QPARAM_ARCHPATH] = archpath

        # ETL Configuration
        if etl:
            params[QPARAM_ETL_NAME] = etl.name
            if etl.args:
                params[QPARAM_ETL_ARGS] = etl.args

        return self._client.get_full_url(self._object_path, params)

    def put_content(self, content: bytes) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_content' instead.

        Puts bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to put as an object.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
        """
        warnings.warn(
            "The 'put_content' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.put_content' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_content(content)

    def put_file(self, path: str or Path) -> Response:
        """
        Deprecated: Use 'ObjectWriter.put_file' instead.

        Puts a local file as an object to a bucket in AIS storage.

        Args:
            path (str or Path): Path to local file

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            ValueError: The path provided is not a valid file
        """
        warnings.warn(
            "The 'put_file' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.put_file' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().put_file(path)

    def get_writer(self) -> ObjectWriter:
        """
        Create an ObjectWriter to write to object contents and attributes.

        Returns:
            An ObjectWriter which can be used to write to an object's contents and attributes.
        """
        return ObjectWriter(self._client, self._object_path, self.query_params)

    # pylint: disable=too-many-arguments
    def promote(
        self,
        path: str,
        target_id: str = "",
        recursive: bool = False,
        overwrite_dest: bool = False,
        delete_source: bool = False,
        src_not_file_share: bool = False,
    ) -> str:
        """
        Promotes a file or folder an AIS target can access to a bucket in AIS storage.
        These files can be either on the physical disk of an AIS target itself or on a network file system
        the cluster can access.
        See more info here: https://aiatscale.org/blog/2022/03/17/promote

        Args:
            path (str): Path to file or folder the AIS cluster can reach
            target_id (str, optional): Promote files from a specific target node
            recursive (bool, optional): Recursively promote objects from files in directories inside the path
            overwrite_dest (bool, optional): Overwrite objects already on AIS
            delete_source (bool, optional): Delete the source files when done promoting
            src_not_file_share (bool, optional): Optimize if the source is guaranteed to not be on a file share

        Returns:
            Job ID (as str) that can be used to check the status of the operation, or empty if job is done synchronously

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            AISError: Path does not exist on the AIS cluster storage
        """
        value = PromoteAPIArgs(
            source_path=path,
            object_name=self.name,
            target_id=target_id,
            recursive=recursive,
            overwrite_dest=overwrite_dest,
            delete_source=delete_source,
            src_not_file_share=src_not_file_share,
        ).as_dict()
        json_val = ActionMsg(action=ACT_PROMOTE, name=path, value=value).dict()

        return self._client.request(
            HTTP_METHOD_POST,
            path=self._bck_path,
            params=self.query_params,
            json=json_val,
        ).text

    def delete(self) -> Response:
        """
        Delete an object from a bucket.

        Returns:
            None

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        return self._client.request(
            HTTP_METHOD_DELETE,
            path=self._object_path,
            params=self.query_params,
        )

    def blob_download(
        self,
        chunk_size: int = None,
        num_workers: int = None,
        latest: bool = False,
    ) -> str:
        """
        A special facility to download very large remote objects a.k.a. BLOBs
        Returns job ID that for the blob download operation.

        Args:
            chunk_size (int): chunk size in bytes
            num_workers (int): number of concurrent blob-downloading workers (readers)
            latest (bool): GET the latest object version from the associated remote bucket

        Returns:
            Job ID (as str) that can be used to check the status of the operation

        Raises:
            aistore.sdk.errors.AISError: All other types of errors with AIStore
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.exceptions.HTTPError: Service unavailable
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
        """
        params = self.query_params.copy()
        value = BlobMsg(
            chunk_size=chunk_size,
            num_workers=num_workers,
            latest=latest,
        ).as_dict()
        json_val = ActionMsg(
            action=ACT_BLOB_DOWNLOAD, value=value, name=self.name
        ).dict()
        return self._client.request(
            HTTP_METHOD_POST, path=self._bck_path, params=params, json=json_val
        ).text

    def append_content(
        self, content: bytes, handle: str = "", flush: bool = False
    ) -> str:
        """
        Deprecated: Use 'ObjectWriter.append_content' instead.

        Append bytes as an object to a bucket in AIS storage.

        Args:
            content (bytes): Bytes to append to the object.
            handle (str): Handle string to use for subsequent appends or flush (empty for the first append).
            flush (bool): Whether to flush and finalize the append operation, making the object accessible.

        Returns:
            handle (str): Handle string to pass for subsequent appends or flush.

        Raises:
            requests.RequestException: "There was an ambiguous exception that occurred while handling..."
            requests.ConnectionError: Connection error
            requests.ConnectionTimeout: Timed out connecting to AIStore
            requests.ReadTimeout: Timed out waiting response from AIStore
            requests.exceptions.HTTPError(404): The object does not exist
        """
        warnings.warn(
            "The 'append_content' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.append_content' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().append_content(content, handle, flush)

    def set_custom_props(
        self, custom_metadata: Dict[str, str], replace_existing: bool = False
    ) -> Response:
        """
        Deprecated: Use 'ObjectWriter.set_custom_props' instead.

        Set custom properties for the object.

        Args:
            custom_metadata (Dict[str, str]): Custom metadata key-value pairs.
            replace_existing (bool, optional): Whether to replace existing metadata. Defaults to False.
        """
        warnings.warn(
            "The 'set_custom_props' method is deprecated and will be removed in a future release. "
            "Please use 'ObjectWriter.set_custom_props' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_writer().set_custom_props(custom_metadata, replace_existing)



"""
Dynamic Batch Sampler for Dynamic Batch Sizing

In scenarios where memory is a constraint, the DynamicBatchSampler
can be used to generate mini-batches that fit within a memory constraint
so that there is a guarantee that each batch fits within memory
while attempting to fit the maximum number of samples in each batch.

Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
"""

import torch
from typing import Iterator, List
from aistore.pytorch.base_map_dataset import AISBaseMapDataset
from logging import getLogger


# Default saturation of a batch needed to not be dropped with drop_last=True
SATURATION_FACTOR = 0.8
logger = getLogger(__name__)


class DynamicBatchSampler(torch.utils.data.Sampler):
    """

    Dynamically adds samples to mini-batch up to a maximum batch size.

    NOTE: Using this sampler with AISBaseMapDatasets that use ObjectGroups
    in their ais_source_lists will be slower than using it with Buckets as
    ObjectGroups will perform one extra API call per object to get size metadata.

    Args:
        data_source (AISBaseMapDataset): Base AIS map-style dataset to sample from to create dynamic mini-batches.
        max_batch_size (float): Maximum size of mini-batch in bytes.
        drop_last (bool, optional): If `True`, then will drop last batch if the batch is not atleast 80% of `max_batch_size`.
            Defaults to `False`.
        allow_oversized_samples (bool, optional): If `True`, then any sample that is larger than the `max_batch_size` will be processed
            in its own min-batch by itself instead of being dropped. Defaults to `False`.
        saturation_factor (float, optional): Saturation of a batch needed to not be dropped with `drop_last=True`. Default is `0.8`.
        shuffle (bool, optional): Randomizes order of samples before calculating mini-batches. Default is `False`.
    """

    def __init__(
        self,
        data_source: AISBaseMapDataset,
        max_batch_size: float,
        drop_last: bool = False,
        allow_oversized_samples: bool = False,
        saturation_factor: float = SATURATION_FACTOR,
        shuffle: bool = False,
    ) -> None:
        super().__init__()
        self._data_source = data_source
        self._max_batch_size = max_batch_size
        self._samples_list = data_source.get_obj_list()
        self._drop_last = drop_last
        self._allow_oversized_samples = allow_oversized_samples
        if not (0 <= saturation_factor <= 1):
            raise ValueError(f"`saturation_factor` must be between 0 and 1")
        self._saturation_factor = saturation_factor
        self._shuffle = shuffle

    def __iter__(self) -> Iterator[List[int]]:
        """
        Returns an iterator containing mini-batches (lists of indices).
        """
        total_mem = 0
        batch = []

        if self._shuffle:
            self._indices = list(
                torch.randperm(len(self))
            )  # randomized list of indices

        # Get sample size for each index, check if there is space in the batch, and yield batches whenever full
        # Calculate spaces in batch non-preemptively
        index = self._get_next_index(-1)
        while index < len(self._samples_list):
            sample = self._samples_list[index]

            if sample.props.size == 0:
                logger.warning(
                    f"Sample {sample.name} cannot be processed as it has a size of 0 bytes"
                )
                index = self._get_next_index(index)
                continue

            if sample.props.size > self._max_batch_size:
                if self._allow_oversized_samples is True:
                    yield [index]
                else:
                    logger.warning(
                        f"Sample {sample.name} cannot be processed as it is larger than the max batch size: {sample.props.size} bytes > {self._max_batch_size} bytes"
                    )

                index = self._get_next_index(index)
                continue

            if total_mem + sample.props.size < self._max_batch_size:
                batch.append(index)
                index = self._get_next_index(index)
                total_mem += sample.props.size
            else:

                if total_mem + sample.props.size == self._max_batch_size:
                    batch.append(index)
                    index = self._get_next_index(index)

                yield batch
                batch = []
                total_mem = 0

        # If batch exists and we are not dropping last or if we are dropping last but the batch is saturated
        # then yield the last batch
        if (batch and not self._drop_last) or (
            self._drop_last
            and total_mem / self._max_batch_size > self._saturation_factor
        ):
            yield batch

    def __len__(self) -> int:
        """
        Returns the total number of samples.
        """
        return len(self._samples_list)

    def _get_next_index(self, index) -> int:
        """
        Get next index from indices if shuffling or otherwise return incremented count.

        Returns:
            index (int): Next index to sample from
        """
        if self._shuffle:
            if len(self._indices) == 0:
                return len(self._samples_list)
            return int(self._indices.pop())
        index += 1
        return index



#
# Copyright (c) 2024-2025, NVIDIA CORPORATION. All rights reserved.
#

import requests

from aistore.sdk.utils import parse_http_error_or_raise
from aistore.sdk.authn.errors import (
    AuthNError,
    ErrClusterNotFound,
    ErrClusterAlreadyRegistered,
    ErrRoleNotFound,
    ErrRoleAlreadyExists,
    ErrUserNotFound,
    ErrUserAlreadyExists,
    ErrUserInvalidCredentials,
)


def parse_authn_error(resp: requests.Response) -> AuthNError:
    """
    Parse raw text into an appropriate AuthNError object.

    Args:
        resp (requests.Response): The response from the AuthN cluster.

    Raises:
        AuthNError: If the error doesn't match any specific conditions.
        ErrClusterNotFound: If the error message indicates a missing cluster.
        ErrRoleNotFound: If the error message indicates a missing role.
        ErrUserNotFound: If the error message indicates a missing user.
        ErrRoleAlreadyExists: If the error message indicates a role already exists.
        ErrUserAlreadyExists: If the error message indicates a user already exists.
        ErrClusterAlreadyRegistered: If the error message indicates the cluster is already registered.
        ErrUserInvalidCredentials: If the error message indicates invalid user credentials.
    """
    exc_class = AuthNError
    err = parse_http_error_or_raise(resp, exc_class)
    status, message = err.status, err.message

    if status == 401:
        if "invalid credentials" in message:
            exc_class = ErrUserInvalidCredentials
    if status == 404:
        if "cluster " in message:
            exc_class = ErrClusterNotFound
        if "role " in message:
            exc_class = ErrRoleNotFound
        if "user " in message:
            exc_class = ErrUserNotFound
    if status == 409:
        if "cluster " in message:
            exc_class = ErrClusterAlreadyRegistered
        if "role " in message:
            exc_class = ErrRoleAlreadyExists
        if "user " in message:
            exc_class = ErrUserAlreadyExists

    return exc_class(status, message)



#
# Copyright (c) 2022-2025, NVIDIA CORPORATION. All rights reserved.
#

import logging
import re
from pathlib import Path
from typing import Callable, Iterator, Optional, Tuple, Type, TypeVar
from urllib.parse import urlparse

import braceexpand
import humanize
import requests
import xxhash

from msgspec import msgpack
from pydantic.v1 import BaseModel, parse_raw_as

from aistore.sdk.const import (
    UTF_ENCODING,
    HEADER_CONTENT_TYPE,
    MSGPACK_CONTENT_TYPE,
    DEFAULT_LOG_FORMAT,
    XX_HASH_SEED,
)
from aistore.sdk.errors import (
    AISError,
    ErrBckNotFound,
    ErrRemoteBckNotFound,
    ErrBckAlreadyExists,
    ErrObjNotFound,
    ErrETLNotFound,
    ErrETLAlreadyExists,
    InvalidBckProvider,
)
from aistore.sdk.provider import Provider

T = TypeVar("T")
MASK = 0xFFFFFFFFFFFFFFFF  # 64-bit mask
# fmt: off
GOLDEN_RATIO = 0x9e3779b97f4a7c15
CONST1 = 0xbf58476d1ce4e5b9
CONST2 = 0x94d049bb133111eb
# fmt: on
ROTATION_BITS = 7


class HttpError(BaseModel):
    """
    Represents an error returned by the API.
    """

    status: int
    message: str = ""
    method: str = ""
    url_path: str = ""
    remote_addr: str = ""
    caller: str = ""
    node: str = ""


def decode_response_text(resp: requests.Response) -> str:
    """
    Convert the response body to a decoded string if it is bytes. Otherwise, return as is.

    Args:
        resp (requests.Response): The response object.

    Returns:
        str: The decoded response text (UTF-8 or ISO-8859-1).
    """
    error_text = resp.text
    if isinstance(error_text, bytes):
        try:
            error_text = error_text.decode(UTF_ENCODING)
        except UnicodeDecodeError:
            error_text = error_text.decode("iso-8859-1")
    return error_text


def parse_http_error_or_raise(
    resp: requests.Response, exc_class: Type[Exception]
) -> HttpError:
    """
    Parse the response body as a HttpError object or raise a custom exception on failure.

    Args:
        resp (requests.Response): The response to parse.
        exc_class (Type[Exception]): The exception class to raise if parsing fails.

    Returns:
        HttpError: The parsed HttpError model.

    Raises:
        exc_class: If the response text cannot be parsed into HttpError.
    """
    error_text = decode_response_text(resp)
    try:
        return parse_raw_as(HttpError, error_text)
    except Exception as exc:
        raise exc_class(resp.status_code, error_text) from exc


def parse_ais_error(resp: requests.Response) -> AISError:
    """
    Parse raw text into an appropriate AISError object.

    Args:
        resp (requests.Response): The response from the AIS cluster.

    Returns:
        AISError: If the error doesn't match any specific conditions.
        ErrBckNotFound: If the error message indicates a missing bucket.
        ErrRemoteBckNotFound: If the error message indicates a missing remote bucket.
        ErrObjNotFound: If the error message indicates a missing object.
        ErrBckAlreadyExists: If the error message indicates a bucket already exists.
        ErrETLAlreadyExists: If the error message indicates an ETL already exists
        ErrETLNotFound: If the error message indicates a missing ETL.
    """
    exc_class = AISError
    err = parse_http_error_or_raise(resp, exc_class)
    status, message = err.status, err.message
    prov, bck, obj = _extract_and_parse_url(message) or (None, None, None)
    if prov:
        try:
            prov = Provider.parse(prov)
        except InvalidBckProvider:
            prov = None

    if status == 404:
        if "etl job" in message:
            exc_class = ErrETLNotFound
        elif prov:
            if obj:
                exc_class = ErrObjNotFound
            elif prov.is_remote() or "@" in bck:
                exc_class = ErrRemoteBckNotFound
            else:
                exc_class = ErrBckNotFound
    elif status == 409:
        if "etl job" in message:
            exc_class = ErrETLAlreadyExists
        elif prov and not obj:
            exc_class = ErrBckAlreadyExists

    return exc_class(status, message)


# pylint: disable=unused-variable
def handle_errors(
    resp: requests.Response,
    parse_error: Callable[[requests.Response], Exception],
) -> None:
    """
    Error handling for requests made to the AIS Client

    Args:
        resp: requests.Response = Response received from the request
        parse_error_fn: Function that processes error text and returns appropriate exceptions.
    """
    try:
        # Will raise HTTPError if status code is 4xx or 5xx
        resp.raise_for_status()
    except requests.exceptions.HTTPError as http_err:
        raise parse_error(resp) from http_err

    # Final status code (post-redirection) wasn't 4xx/5xx. Could be 1xx or 3xx.
    raise parse_error(resp)


def probing_frequency(dur: int) -> float:
    """
    Given a timeout, return an interval to wait between retries

    Args:
        dur: Duration of timeout

    Returns:
        Frequency to probe
    """
    freq = min(dur / 8.0, 1.0)
    freq = max(dur / 64.0, freq)
    return max(freq, 0.1)


def read_file_bytes(filepath: str) -> bytes:
    """
    Given a filepath, read the content as bytes
    Args:
        filepath: Existing local filepath

    Returns: Raw bytes
    """
    with open(filepath, "rb") as reader:
        return reader.read()


def _check_path_exists(path: Path) -> None:
    if not path.exists():
        raise ValueError(f"Path: {path} does not exist")


def validate_file(path: str or Path) -> None:
    """
    Validate that a file exists and is a file
    Args:
        path (str or Path): Path to validate
    Raises:
        ValueError: If path does not exist or is not a file
    """
    if isinstance(path, str):
        path = Path(path)
    _check_path_exists(path)
    if not path.exists():
        raise ValueError(f"Path: {path} does not exist")
    if not path.is_file():
        raise ValueError(f"Path: {path} is a directory, not a file")


def validate_directory(path: str or Path) -> None:
    """
    Validate that a directory exists and is a directory
    Args:
        path (str or Path): Path to validate
    Raises:
        ValueError: If path does not exist or is not a directory
    """
    if isinstance(path, str):
        path = Path(path)
    _check_path_exists(path)
    if not path.is_dir():
        raise ValueError(f"Path: {path} is a file, not a directory")


def get_file_size(file: Path) -> str:
    """
    Get the size of a file and return it in human-readable format
    Args:
        file: File to read

    Returns:
        Size of file as human-readable string

    """
    return (
        humanize.naturalsize(file.stat().st_size) if file.stat().st_size else "unknown"
    )


def expand_braces(template: str) -> Iterator[str]:
    """
    Given a string template, apply bash-style brace expansion to return a list of strings
    Args:
        template: Valid brace expansion input, e.g. prefix-{0..10..2}-gap-{11..15}-suffix

    Returns:
        Iterator of brace expansion output

    """
    # pylint: disable = fixme
    # TODO Build custom expansion to validate consistent with cmn/cos/template.go TemplateRange
    return braceexpand.braceexpand(template)


def decode_response(
    res_model: Type[T],
    resp: requests.Response,
) -> T:
    """
    Parse response content from the cluster into a Python class,
     decoding with msgpack depending on content type in header

    Args:
        res_model (Type[T]): Resulting type to which the response should be deserialized
        resp (Response): Response from the AIS cluster

    """
    if resp.headers.get(HEADER_CONTENT_TYPE) == MSGPACK_CONTENT_TYPE:
        return msgpack.decode(resp.content, type=res_model)
    return parse_raw_as(res_model, resp.text)


def parse_url(url: str) -> Tuple[str, str, str]:
    """
    Parse AIS URLs for bucket and object names.

    Args:
        url (str): Complete URL of the object (e.g., "ais://bucket1/file.txt")

    Returns:
        Tuple[str, str, str]: Provider, bucket name, and object name
    """
    parsed_url = urlparse(url)
    path = parsed_url.path.lstrip("/")
    return parsed_url.scheme, parsed_url.netloc, path


def _extract_and_parse_url(msg: str) -> Optional[Tuple[str, str, str]]:
    url_pattern = r"[a-z0-9]+://[A-Za-z0-9@._/-]+"
    match = re.search(url_pattern, msg)
    if match:
        return parse_url(match.group())
    return None


def get_logger(name: str, log_format: str = DEFAULT_LOG_FORMAT):
    """
    Create or retrieve a logger with the specified configuration.

    Args:
        name (str): The name of the logger.
        log_format (str, optional): Logging format.

    Returns:
        logging.Logger: Configured logger instance.
    """
    logger = logging.getLogger(name)
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(log_format))
        logger.addHandler(handler)
    logger.propagate = False
    return logger


# Translated from:
# http://xoshiro.di.unimi.it/xoshiro256starstar.c
# Scrambled Linear Pseudorandom Number Generators
# David Blackman, Sebastiano Vigna
# https://arxiv.org/abs/1805.01407
# http://www.pcg-random.org/posts/a-quick-look-at-xoshiro256.html
def xoshiro256_hash(seed: int) -> int:
    """
    Xoshiro256-inspired hash function with 64-bit overflow behavior.
    """
    z = (seed + GOLDEN_RATIO) & MASK
    z = (z ^ (z >> 30)) * CONST1 & MASK
    z = (z ^ (z >> 27)) * CONST2 & MASK
    z = (z ^ (z >> 31)) + GOLDEN_RATIO & MASK
    z = (z ^ (z >> 30)) * CONST1 & MASK
    z = (z ^ (z >> 27)) * CONST2 & MASK
    z = (z ^ (z >> 31)) * 5 & MASK
    rotated = ((z << ROTATION_BITS) | (z >> (64 - ROTATION_BITS))) & MASK
    return (rotated * 9) & MASK


def get_digest(name: str) -> int:
    """
    Get the xxhash digest of a given string.
    """
    return xxhash.xxh64(seed=XX_HASH_SEED, input=name.encode("utf-8")).intdigest()
